# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"max_grad": "00_core.ipynb",
         "semiring": "00_core.ipynb",
         "Log": "00_core.ipynb",
         "Max": "00_core.ipynb",
         "grad": "00_core.ipynb",
         "SequenceDist": "00_core.ipynb",
         "to_np": "01_utils.ipynb",
         "zero_grad": "01_utils.ipynb",
         "float64": "01_utils.ipynb",
         "compare_fwd_bwd": "01_utils.ipynb",
         "timed": "01_utils.ipynb",
         "benchmark_fwd_bwd": "01_utils.ipynb",
         "report": "01_utils.ipynb",
         "add_checks": "01_utils.ipynb",
         "load_cupy_func": "01_utils.ipynb",
         "load_cupy_module": "01_utils.ipynb",
         "device": "06_Sparse.ipynb",
         "generate_sample_inputs": "04_CTC_loss_simple.ipynb",
         "loss_pytorch": "02_CTC_loss.ipynb",
         "interleave_blanks": "02_CTC_loss.ipynb",
         "prepare_inputs": "03_CTC_loss_banded.ipynb",
         "loss_basic": "03_CTC_loss_banded.ipynb",
         "fwd_bwd": "03_CTC_loss_banded.ipynb",
         "masked_grad": "02_CTC_loss.ipynb",
         "loss_py": "03_CTC_loss_banded.ipynb",
         "loss_cupy": "03_CTC_loss_banded.ipynb",
         "cupy_funcs": "06_Sparse.ipynb",
         "viterbi_alignments": "04_CTC_loss_simple.ipynb",
         "soft_alignments": "04_CTC_loss_simple.ipynb",
         "loss_direct_cupy": "02_CTC_loss.ipynb",
         "Prob": "02_CTC_loss.ipynb",
         "window_start_positions": "03_CTC_loss_banded.ipynb",
         "loss_masked": "03_CTC_loss_banded.ipynb",
         "logZ_fwd": "04_CTC_loss_simple.ipynb",
         "dot": "04_CTC_loss_simple.ipynb",
         "LogZ": "05_Normalisation.ipynb",
         "logZ_py": "05_Normalisation.ipynb",
         "mean": "04_CTC_loss_simple.ipynb",
         "logZ_cupy": "04_CTC_loss_simple.ipynb",
         "cupy_funcs_loop": "04_CTC_loss_simple.ipynb",
         "generate_test_example": "05_Normalisation.ipynb",
         "fused_batch_Mv": "05_Normalisation.ipynb",
         "LogZViterbi": "05_Normalisation.ipynb",
         "logz": "05_Normalisation.ipynb",
         "Mv_scan_py": "06_Sparse.ipynb",
         "logZ_scan_py": "06_Sparse.ipynb",
         "ctc_loss_scan_py": "06_Sparse.ipynb",
         "Mv_scan_cupy": "06_Sparse.ipynb",
         "logZ_scan": "06_Sparse.ipynb",
         "ctc_loss_scan": "06_Sparse.ipynb",
         "cupy_func": "06_Sparse.ipynb",
         "logZ_fwd_cupy": "06_Sparse.ipynb",
         "fwd_scores_cupy": "06_Sparse.ipynb",
         "bwd_scores_cupy": "06_Sparse.ipynb",
         "logZ": "06_Sparse.ipynb",
         "logZ_bwd_cupy": "06_Sparse.ipynb",
         "ctc_loss": "06_Sparse.ipynb"}

modules = ["core.py",
           "utils.py",
           "ctc.py",
           "ctc_banded.py",
           "ctc_simple.py",
           "normalisation.py",
           "sparse.py"]

doc_url = "https://davidcpage.github.io/seqdist/"

git_url = "https://github.com/davidcpage/seqdist/tree/master/"

def custom_doc_links(name): return None
